/* Name: main.c
 * Author: <insert your name here>
 * Copyright: <insert your copyright message here>
 * License: <insert your license reference here>
 */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/interrupt.h>  
#include <avr/sleep.h>
#include <avr/eeprom.h>

//number of pixels in the output
#define ARRAY_LENGTH 100
	
//number of characters in output
#define MESSAGE_LENGTH 12

//time until unit stops trying to display text and lights up indicator LED
#define displayRefreshTimeout 5000

//pixeldata content of each letter A-Z
const uint8_t alphabet[242] PROGMEM = {0x00,0x00,0x00,0x00,0x00,0x00,0x3f,0x7f,0xcc,0xcc,0x7f,0x3f,0xff,0xff,0xdb,0xdb,0xff,0x66,0x3c,0x7e,0xe7,0xc3,0xc3,0x66,0xff,0xff,0xc3,0xe7,0x7e,0x3c,0xff,0xff,0xdb,0xdb,0xc3,0xff,0xff,0xd8,0xd8,0xc0,0x3c,0x7e,0xe7,0xc3,0xcb,0xcf,0x4e,0xff,0xff,0x18,0x18,0xff,0xff,0xc3,0xc3,0xff,0xff,0xc3,0xc3,0x0c,0xc3,0xc3,0xc3,0xfe,0xfc,0xff,0xff,0x18,0x3e,0x67,0x03,0xff,0xff,0x03,0x03,0x03,0xff,0xff,0x60,0x30,0x60,0xff,0xff,0xff,0xff,0x60,0x38,0x0c,0xff,0xff,0x3c,0x7e,0xc3,0xc3,0x7e,0x3c,0xff,0xff,0xcc,0xcc,0x78,0x30,0x3c,0x7e,0xc3,0xc3,0xfe,0x3d,0xff,0xff,0xcc,0xcc,0x7e,0x37,0x76,0xf3,0x93,0xcf,0x4e,0xc0,0xc0,0xff,0xff,0xc0,0xc0,0xfe,0xff,0x03,0x03,0xff,0xfe,0xfc,0xfe,0x03,0x03,0xfe,0xfc,0xff,0xff,0x06,0x0c,0x06,0xff,0xff,0xc3,0x7e,0x18,0x18,0x7e,0xc3,0xc0,0x60,0x3f,0x3f,0x60,0xc0,0xc3,0xc7,0xcf,0xf3,0xe3,0xc3,0x23,0x63,0xff,0xff,0x03,0x03,0x43,0xc7,0xcf,0xfb,0x73,0x42,0xd3,0xd3,0xff,0x6e,0xf8,0xf8,0x18,0x18,0xff,0xff,0xfa,0xdb,0xdb,0xdf,0xce,0x7e,0xff,0xd3,0xdf,0x5e,0xc7,0xcf,0xd8,0xf0,0xe0,0x6e,0xff,0x93,0x93,0xff,0x6e,0x70,0xfa,0xdb,0xdb,0x7e,0x3c,0x7e,0xff,0xcb,0xd3,0xff,0x7e,0xfb,0xfb,0x60,0xe0,0xcb,0xdb,0xf0,0x70,0x14,0x14,0x7f,0x14,0x7f,0x14,0x14,0x03,0x03,0x66,0x66,0x64,0x66,0x02,0x02,0x66,0x64};

//how many pixels wide each letter is.
const uint8_t letterlen[43] PROGMEM = {6,6,6,6,6,5,5,7,6,6,6,6,5,7,7,6,6,6,6,5,6,6,6,7,6,6,6,6,5,5,6,5,5,5,6,6,6,2,6,7,2,2,6};
	
//placement of the beginning of each letter (basically a running sum of letterlen)
const uint8_t letterpos[43] PROGMEM = {0,6, 12, 18, 24, 30, 35, 40, 47, 53, 59, 65, 71, 76, 83, 90, 96, 102, 108, 114, 119, 125, 131, 137, 144, 150, 156, 162, 168, 173, 178, 184, 189, 194, 199, 205, 211, 217, 219, 225,232,234,236};

//Default messages
const uint8_t defaultMessages[8][MESSAGE_LENGTH] PROGMEM ={{8,5,12,12,15,0,0,0,0,0,0,0},\
                                                    {7,15,15,4,2,25,5,0,0,0,0,0},\
                                                    {0,0,0,0,0,0,0,0,0,0,0,0},\
                                                    {0,0,0,0,0,0,0,0,0,0,0,0},\
                                                    {0,0,0,0,0,0,0,0,0,0,0,0},\
                                                    {0,0,0,0,0,0,0,0,0,0,0,0},\
                                                    {0,0,0,0,0,0,0,0,0,0,0,0},\
                                                    {0,0,0,0,0,0,0,0,0,0,0,0},\
                                                    {0,0,0,0,0,0,0,0,0,0,0,0}};

//location of current letter being edited
volatile uint8_t letterCursor;

//which of the 8 messages is currently being displayed/edited
volatile uint8_t messageNumber = 0;
	
//ignore the first few pulses so you don't prematurely jump into display mode.
volatile uint8_t ignoreShakes = 2;

//set when cursor is moved, cleared when letter is changed
volatile uint8_t cursorMoved=1;
	
//characters in output
volatile uint8_t outputText[8][MESSAGE_LENGTH]={};


//framebuffer of current thing to display
volatile uint8_t outputarray[100]={0};



//duration of current pixel output
volatile uint8_t count;

//duration of a complete wave cycle
volatile uint32_t totalcount;

//the pixel row currently being displayed
volatile uint8_t currentrow=0;

//time length of each row
volatile uint8_t rowcount;

//amount of time between accelerometer signal and start of display.
volatile uint32_t timeout;

//location of "cursor" that is editing the array when adding/removing letters
volatile uint8_t arraypos = 0;

//counts down to when the device should sleep
volatile uint32_t sleepcount = 0;               

//setting to 1 sleeps the device on its next time through the main loop.
volatile uint32_t sleepme = 0;

//enable editing
volatile uint8_t editMode = 0;

//draw cursor on next cycle
volatile uint8_t drawCursor = 0;

//indicate that image should be redrawn next time through the main loop
//do this every time you change something
volatile uint8_t redraw = 1;

void loadFromEEPROM()
{
    uint8_t i;
    uint8_t j;
    for (i=0;i<8;i++)
    {
        for (j=0;j<MESSAGE_LENGTH;j++)
        {
            outputText[i][j]=eeprom_read_byte((uint8_t*)((i*MESSAGE_LENGTH)+j));
        }
    }
}

void saveToEEPROM()
{
    uint8_t i;
    uint8_t j;
    for (i=0;i<8;i++)
    {
        for (j=0;j<MESSAGE_LENGTH;j++)
        {
            eeprom_write_byte((uint8_t*)((i*MESSAGE_LENGTH)+j),outputText[i][j]);
        }
    }
}

void initEEPROM()
{
    uint8_t i;
    uint8_t j;
    //reset EEPROM
    if (!(PINB&4))
    {
        for (i=0;i<8;i++)
        {
            for (j=0;j<MESSAGE_LENGTH;j++)
            {
                eeprom_write_byte((uint8_t*)((i*MESSAGE_LENGTH)+j),pgm_read_byte(&(defaultMessages[i][j])));
            }
        }
    }
    loadFromEEPROM();
}

void initPorts()
{
	//PA0-6 and PB0 are LED outputs
	//PB2 is button input
	//PA7 is input from accelerometer
	//PB1controls accelerometer enable
	DDRA = 0X7F;
        DDRB = 0X03;
        PORTA = 0x7F;
        PORTB = 0x03;
}

void disablePorts()
{
	DDRA = 0X00;
        DDRB = 0X00;
        PORTA = 0x00;
        PORTB = 0x00;
}

//totally clear out the rest of the frame buffer
void resetOutputArray()
{
    while (arraypos<ARRAY_LENGTH)
    {
        outputarray[arraypos]=0xff;
        arraypos++;
    }
    arraypos = 0;
    
}



//add letter to end
void addLetter(uint8_t letter)
{
    uint8_t currentletterlength = pgm_read_byte(&letterlen[letter]);
    uint8_t currentletterposition = pgm_read_byte(&letterpos[letter]);
    uint8_t j;
    for (j=0;j<currentletterlength;j++)
    {
        outputarray[arraypos]=~pgm_read_byte(&alphabet[currentletterposition+j]);
        arraypos++;
    }
    
    //gaps after letters
    outputarray[arraypos]=0xff; 
    arraypos++;
    outputarray[arraypos]=0xff;
    arraypos++;
}

//add cursor block
void addCursor(uint8_t letter)
{
    uint8_t currentletterlength = pgm_read_byte(&letterlen[letter]);
    uint8_t j;
    for (j=0;j<currentletterlength;j++)
    {
	if (letter!=0)
		outputarray[arraypos]=~(0x1);
	else
		outputarray[arraypos]=0xFE; //for spaces, make special character
        arraypos++;
    }
    
    //gaps after letters
    outputarray[arraypos]=0xff; 
    arraypos++;
    outputarray[arraypos]=0xff;
    arraypos++;
}

//generate output from text
void genOutputArray()
{
	
	uint8_t i;
	arraypos=1;
	for(i=0;i<MESSAGE_LENGTH;i++)
	{
		if ((i==letterCursor)&&(drawCursor)&&(editMode))
			addCursor(outputText[messageNumber][i]);
		else
			addLetter(outputText[messageNumber][i]);
	}
	resetOutputArray();
}

//advance cursor
void longPush()
{
	if (editMode)
	{
        cursorMoved=1;
		if (letterCursor<(MESSAGE_LENGTH-1))
			letterCursor++;
		else
			letterCursor=0;
	}
}

//enter/exit edit mode
void reallyLongPush()
{
	redraw=1;
	if (editMode)
    {
        saveToEEPROM();
		editMode=0;
        
    }
	else
    {
		editMode=1;
        cursorMoved=1;
    }
}

void deletePattern(uint8_t delPattern)
{
	uint8_t i;
	for (i=0;i<MESSAGE_LENGTH;i++)
	{
		outputText[delPattern][i]=0;
	}
}

//clear pattern
void reallyReallyLongPush()
{
	if (editMode)
	{
		deletePattern(messageNumber);
		letterCursor=0;
	}
	else
		reallyLongPush(); //if user activates reallyReallyLongPush() from non-edit mode, they probably meant to invoke
					//reallyLongPush() and just held on too long.
	
}


	
//advance letter/pattern
void shortPush()
{
	if (editMode)
	{
        //if you just moved the cursor on to this letter, start from beginning of alphabet
        if (cursorMoved)
        {
            cursorMoved=0;
            outputText[messageNumber][letterCursor] = 1;
        }
        else if (outputText[messageNumber][letterCursor]<42)
			outputText[messageNumber][letterCursor] = outputText[messageNumber][letterCursor]+1;
		else
			outputText[messageNumber][letterCursor] = 0;
		redraw=1;
	}
	else
	{
        letterCursor=0;
		if (messageNumber<7)
		{
			messageNumber++;
			redraw=1;
		}
		else
		{
			messageNumber=0;
			redraw=1;
		}
	}
	
}

//print byte to display (0 is on, 1 is off)
void printByte(uint8_t byte)
{
	PORTA = ((((byte&0xF8)>>3)|((byte&0x3)<<5))&(0x7F));
	PORTB = (((byte&0b100)>>2)|0x2);
}

//print item from current row of display buffer to LEDs
void printrow(uint8_t row)
{
	printByte(outputarray[row]);
	return;
}



//blink all the lights once
void blinkLights()
{
        printByte(0x00);
	_delay_ms(10);
	printrow(currentrow);
	redraw=1;
}


int main(void)
{
	
	//initialize all GPIO pins
	initPorts();
    
    //initialize memory
    initEEPROM();
    
	//initialize all public variables
	totalcount = 0;
	count = 0;
	currentrow = 0;
	rowcount = 0;
	timeout = 0;
	
	//first element of output array is always blank.
	//it will be what's displayed during the "back" swipe
	outputarray[0]=0xff;

	//configure sleep settings and timer interrupts
	MCUCR = 0X10; //POWER-DOWN sleep mode //low-level interrupt
	GIMSK = 0x50; //INT0 ENABLED. Pin Change Interrupt [0:7] enabled
	PCMSK0 = 0X80;//enable interrupt PCINT7
    
    
	//Timer0 interrupt
	OCR0A = 0x10; //how high you count
	TCCR0A = 0x02; //no output, mode 4 (clear timer on compare)
	TCCR0B = 0x03; //no prescaling.
	TIMSK0 = 0x02; //enable compare match interrupt.
	TCNT0 = 0;
    
	GIFR = 0x10;	//clear interrupt flag
	sei();
    
	;
	
	letterCursor = 0x00;
    
    for (;;)
    {
	    if (redraw)
	    {
		    redraw = 0;
		    genOutputArray();
	    }
        //button is pressed
        if (!(PINB&4))
        {
	    //keep device from sleeping
            sleepcount=0;
		
            uint8_t buttonPushTime=0;
	    //wait for user to let go
            while (!(PINB&4))
            {
		    //cap buttonPushTime to keep it from rolling over
		if (buttonPushTime<200)
		{
			buttonPushTime++;
			//indicate that user has held long enough to invoke longPush()
			if ((buttonPushTime==20)&editMode)
			{
				blinkLights();
			}
			//indicate that user has invoked reallyLongPush()
			if (buttonPushTime==50)
			{
				blinkLights();
				_delay_ms(5);
				blinkLights();
			}
			//indicate that user has invoked reallyReallyLongPush()
			if ((buttonPushTime==100)&editMode)
			{
				blinkLights();
				_delay_ms(5);
				blinkLights();
				_delay_ms(5);
				blinkLights();
			}
		}
		else
		{
		}
                _delay_ms(1);
            }
            //ignore really short presses for debouncing
            if (buttonPushTime<5)
                ;
            else if (buttonPushTime<20)
                shortPush();
            else if (buttonPushTime<50)
                longPush();
            else if (buttonPushTime<100)
                reallyLongPush();
            else
                reallyReallyLongPush();
	    
        }
        
	
	//if device has been told to sleep, sleep
        if (sleepme)
        {
            sleep_enable();
	    disablePorts();
            PCMSK0 = 0X00;//disable interrupt PCINT7
            sleep_cpu();
            PCMSK0 = 0X80;//enable interrupt PCINT7
            initPorts();
            sleep_disable();
	    while (!(PINB&4)) //wait for user to let go of button so button press only wakes device and nothing else
            {}
	    editMode=0;
            saveToEEPROM();
	    sleepcount=0;
            redraw = 1; //if you go to sleep with the cursor shown, want to remove it when you awake
            sleepme=0;
            totalcount=0;
            
        }
        
    }
    
}





ISR(TIM0_COMPA_vect)
{
        //if device hasn't been moved for a very long time, put it to sleep
        if (sleepcount<250000)
        {
            sleepcount++;
        }
        else
        {
            sleepme=1;
        }
    
	
    if (totalcount<displayRefreshTimeout)
		totalcount++;
    else
	{
		//enable the shake buffer only right when the LED comes on
		if (totalcount==displayRefreshTimeout)	
		{
			ignoreShakes=2;
			totalcount++; //increment just so ignoreShakes doesn't get called again
		}

			printByte(~(1<<messageNumber));
	}
    
    //don't do anything if you're still in the timeout stage before the message
    if (totalcount>timeout)
    {
	    //keep ticking until you reach the end of this row
        if (count < rowcount)
        {
            count++;
        }
        else
        {
            count = 0;
		//move to next row
            if (currentrow < (ARRAY_LENGTH-1))
            {
                currentrow++;
                printrow(currentrow);
            }
        }
    }
	
		
}

//interrupts every time the accelerometer gets a ping
ISR (PCINT0_vect)
{	
	
	
    uint32_t mastercount;
	if (ignoreShakes>0)
		ignoreShakes--;
	else
	{
		//only concern yourself with high-low transitions
		//hand motion is asymmetrical, so you can only update on swing in one direction
		if (!(PINA&0X80))
		{
			if (editMode)
			{
				if (drawCursor)
					drawCursor=0;
				else
					drawCursor=1;
				redraw=1;
			}
	
			currentrow = 0;
			printrow(currentrow);
			mastercount = totalcount>>1; //allow time for forward and return trip.
        
			//some arbitrary percentage of the total cycle is the timeout.  This allows
			//for the start of the cycle to be at the end of a wave, not in the middle.
			timeout = (mastercount>>1)+(mastercount>>3);
	
			//divy up the amount of time per cycle by the number of rows you hope to display    
			//subtract the timeout to take care of the beginning, and the (mastercount>>4) to tighten up the end a bit.
			rowcount = (mastercount-(mastercount>>4)-(timeout>>1))/(ARRAY_LENGTH);
        
			totalcount=0;
	    
			count=0;
		}
	}
    
	//clear interrupt flag in case there were any spurious interrupts during this vector
	GIFR = (1<<PCIF0);

}



ISR (INT0_vect)
{
}